"""Added importHash

Peek Plugin Database Migration Script

Revision ID: 56d805072d2a
Revises: 3a9cddaf05e5
Create Date: 2017-10-15 12:05:26.795359

"""

# revision identifiers, used by Alembic.
revision = "56d805072d2a"
down_revision = "3a9cddaf05e5"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "InternalGroup",
        sa.Column("importHash", sa.String(length=50), nullable=True),
        schema="core_user",
    )
    op.create_index(
        "idx_InternalGroupTable_importHash",
        "InternalGroup",
        ["importHash"],
        unique=False,
        schema="core_user",
    )
    op.add_column(
        "InternalUser",
        sa.Column("importHash", sa.String(length=50), nullable=True),
        schema="core_user",
    )
    op.create_index(
        "idx_InternalUserTable_importHash",
        "InternalUser",
        ["importHash"],
        unique=False,
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_InternalUserTable_importHash",
        table_name="InternalUser",
        schema="core_user",
    )
    op.drop_column("InternalUser", "importHash", schema="core_user")
    op.drop_index(
        "idx_InternalGroupTable_importHash",
        table_name="InternalGroup",
        schema="core_user",
    )
    op.drop_column("InternalGroup", "importHash", schema="core_user")
    # ### end Alembic commands ###



"""add logged in groups

Peek Plugin Database Migration Script

Revision ID: 2e6fbe116ff9
Revises: d03db0fcb600
Create Date: 2024-03-28 06:54:04.846124

"""

# revision identifiers, used by Alembic.
revision = "2e6fbe116ff9"
down_revision = "d03db0fcb600"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        table_name="UserLoggedIn",
        column=sa.Column("loggedInWithGroups", sa.String(), nullable=True),
        schema="core_user",
    )


def downgrade():
    op.drop_column("UserLoggedIn", "loggedInWithGroups", schema="core_user")



import logging
import os

from celery import Celery
from jsoncfg.value_mappers import require_string
from peek_core_user._private.server.controller.AdminAuthController import (
    AdminAuthController,
)
from sqlalchemy import MetaData, text

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.controller.metrics_controller import (
    MetricsController,
)
from peek_plugin_base.server.PluginLogicEntryHookABC import (
    PluginLogicEntryHookABC,
)
from peek_plugin_base.server.PluginServerStorageEntryHookABC import (
    PluginServerStorageEntryHookABC,
)
from peek_plugin_base.server.PluginServerWorkerEntryHookABC import (
    PluginServerWorkerEntryHookABC,
)
from peek_core_user._private.server.AdminTupleDataObservable import (
    makeAdminTupleDataObservableHandler,
)
from peek_core_user._private.server.ClientTupleActionProcessor import (
    makeTupleActionProcessorHandler,
)
from peek_core_user._private.server.ClientTupleDataObservable import (
    makeTupleDataObservableHandler,
)
from peek_core_user._private.server.admin_backend import (
    makeAdminBackendHandlers,
)
from peek_core_user._private.server.api.UserApi import UserApi
from peek_core_user._private.server.controller.ImportController import (
    ImportController,
)
from peek_core_user._private.server.controller.LoginLogoutController import (
    LoginLogoutController,
)
from peek_core_user._private.server.controller.MainController import (
    MainController,
)
from peek_core_user._private.tuples.LoggedInUserStatusTuple import (
    LoggedInUserStatusTuple,
)
from peek_core_user.server.UserApiABC import UserApiABC
from peek_plugin_base.storage.DbConnection import DbConnection

logger = logging.getLogger(__name__)


class PluginLogicEntryHook(
    PluginLogicEntryHookABC,
    PluginServerStorageEntryHookABC,
    PluginServerWorkerEntryHookABC,
):
    def __init__(self, *args, **kwargs):
        PluginLogicEntryHookABC.__init__(self, *args, **kwargs)
        self._userApi = None

        self._handlers = []

    def _migrateStorageSchema(self, metadata: MetaData) -> None:
        """Migrate Storage Schema

        Rename the schema

        """

        relDir = self._packageCfg.config.storage.alembicDir(require_string)
        alembicDir = os.path.join(self.rootDir, relDir)
        if not os.path.isdir(alembicDir):
            raise NotADirectoryError(alembicDir)

        dbConn = DbConnection(
            dbConnectString=self.platform.dbConnectString,
            metadata=metadata,
            alembicDir=alembicDir,
            enableCreateAll=False,
        )

        # Rename the plugin schema to core.
        renameToCoreSql = text(
            """
            DO $$
            BEGIN
                IF EXISTS(
                    SELECT schema_name
                      FROM information_schema.schemata
                      WHERE schema_name = 'pl_user'
                  )
                THEN
                  EXECUTE ' DROP SCHEMA IF EXISTS core_user CASCADE ';
                  EXECUTE ' ALTER SCHEMA pl_user RENAME TO core_user ';
                END IF;
            END
            $$;
        """
        )

        dbSession = dbConn.ormSessionCreator()
        try:
            dbSession.execute(renameToCoreSql)
            dbSession.commit()

        finally:
            dbSession.close()
            dbConn.dbEngine.dispose()

        PluginServerStorageEntryHookABC._migrateStorageSchema(self, metadata)

    def load(self) -> None:
        from peek_core_user._private.tuples import loadPrivateTuples

        loadPrivateTuples()

        from peek_core_user._private.storage.DeclarativeBase import (
            loadStorageTuples,
        )

        loadStorageTuples()

        from peek_core_user.tuples import loadPublicTuples

        loadPublicTuples()

        # ----------------
        # Setup our API
        self._userApi = UserApi()

        logger.debug("loaded")

    def start(self):
        """Start

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        metricsController = MetricsController(
            writeDir=self.platform.metricsWriteDirectory,
            ormSessionCreator=self.dbSessionCreator,
        )
        metricsController.start()
        self._handlers.append(metricsController)

        # ----------------
        # Setup the APIs
        deviceApi: DeviceApiABC = self.platform.getOtherPluginApi(
            "peek_core_device"
        )
        emailApi = self.platform.getOtherPluginApi("peek_core_email")

        # ----------------
        # Import Controller
        importController = ImportController()
        self._handlers.append(importController)

        # ----------------
        # Login / Logout Controller
        loginLogoutController = LoginLogoutController(
            deviceApi, emailApi, self.dbSessionCreator
        )
        self._handlers.append(loginLogoutController)

        # ----------------
        # Login / Logout Controller
        adminAuthController = AdminAuthController(self.dbSessionCreator)
        self._handlers.append(adminAuthController)

        # ----------------
        # Setup our API
        self._userApi.setStartValues(
            self.dbSessionCreator,
            deviceApi,
            importController,
            loginLogoutController,
            adminAuthController,
        )

        # ----------------
        # Main Controller
        mainController = MainController(self.dbSessionCreator, self._userApi)
        self._handlers.append(mainController)

        # ----------------
        # Client Tuple Observable
        clientTupleObservable = makeTupleDataObservableHandler(
            self.dbSessionCreator, self._userApi
        )
        self._handlers.append(clientTupleObservable)

        # ----------------
        # Admin Tuple Observable
        adminTupleObservable = makeAdminTupleDataObservableHandler(
            self.dbSessionCreator, deviceApi, self._userApi
        )
        self._handlers.append(adminTupleObservable)

        # ----------------
        # Setup controllers.
        loginLogoutController.setup(
            clientTupleObservable,
            adminTupleObservable,
            self._userApi.fieldHookApi,
            self._userApi.infoApi,
        )
        importController.setTupleObserver(clientTupleObservable)

        # Make the admin observable send an update when device online / offline
        # state changes occur
        deviceApi.deviceOnlineStatus().subscribe(
            lambda _: adminTupleObservable.notifyOfTupleUpdateForTuple(
                LoggedInUserStatusTuple.tupleType()
            )
        )

        # ----------------
        # Setup the Action Processor
        self._handlers.append(makeTupleActionProcessorHandler(mainController))

        # ----------------
        # Setup admin backend
        self._handlers.extend(
            makeAdminBackendHandlers(
                clientTupleObservable, self.dbSessionCreator
            )
        )

        logger.debug("started")

    def stop(self):
        for handler in self._handlers:
            handler.shutdown()

        if self._userApi:
            self._userApi.shutdown()

        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")

    @property
    def publishedServerApi(self) -> UserApiABC:
        return self._userApi

    @property
    def dbMetadata(self):
        from peek_core_user._private.storage import DeclarativeBase

        return DeclarativeBase.metadata

    ###### Implement PluginServerWorkerEntryHookABC



import logging

from twisted.internet.defer import Deferred

from peek_core_user._private.server.controller.ImportController import (
    ImportController,
)
from peek_core_user.server.UserImportApiABC import UserImportApiABC

logger = logging.getLogger(__name__)


class UserImportApi(UserImportApiABC):
    def __init__(self):
        self._importController = None

    def setStartValues(self, importController: ImportController):
        self._importController = importController

    def shutdown(self):
        self._importController = None

    def importInternalUsers(
        self, importHash: str, usersEncodedPayload: bytes
    ) -> Deferred:
        return self._importController.importInternalUsers(
            importHash, usersEncodedPayload
        )

    def importInternalGroups(
        self, importHash: str, groupsEncodedPayload: bytes
    ) -> Deferred:
        return self._importController.importInternalGroups(
            importHash, groupsEncodedPayload
        )



"""update user group names

Peek Plugin Database Migration Script

Revision ID: 94d5c124461d
Revises: aa692f37c14d
Create Date: 2024-04-05 15:15:41.462937

"""

# revision identifiers, used by Alembic.
revision = "94d5c124461d"
down_revision = "aa692f37c14d"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    op.execute(
        """
    UPDATE "core_user"."SettingProperty" SET 
    "char_value"='peek-admin-app-login' WHERE "char_value"='peek-admin-login';
    """
    )

    op.execute(
        """
    UPDATE "core_user"."SettingProperty" SET 
    "char_value"='peek-field-app-login' WHERE "char_value"='peek-mobile-login';
    """
    )

    op.execute(
        """
        UPDATE "core_user"."InternalGroup" SET "groupName"='peek-admin-app-login'
        WHERE "groupName"='peek-admin-login';
        """
    )

    op.execute(
        """
        UPDATE "core_user"."InternalGroup" SET "groupName"='peek-field-app-login'
        WHERE "groupName"='peek-mobile-login';
        """
    )


def downgrade():
    pass



from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC

from peek_core_user._private.PluginNames import userPluginTuplePrefix


@addTupleType
class UserLoginOtpAction(TupleActionABC):
    __tupleType__ = userPluginTuplePrefix + "UserLoginOtpAction"

    userName: str = TupleField()



import logging

from twisted.internet.defer import Deferred

from peek_core_user._private.server.controller.AdminAuthController import (
    AdminAuthController,
)
from peek_core_user.server.UserAdminAuthApiABC import UserAdminAuthApiABC

logger = logging.getLogger(__name__)


class UserAdminAuthApi(UserAdminAuthApiABC):
    def __init__(self):
        self._adminAuthController = None

    def setStartValues(self, adminAuthController: AdminAuthController):
        self._adminAuthController = adminAuthController

    def shutdown(self):
        self._adminAuthController = None

    def check(self, username: str, password: str) -> Deferred:
        return self._adminAuthController.check(username, password)



from datetime import datetime

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class UserLoginOtpResponseTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserLoginOtpResponseTuple"

    STATUS_OTP_REQUEST_ACCEPTED = 0
    STATUS_OTP_REQUEST_REJECTED = 1

    otpRequestStatus: int = TupleField()
    userName: str = TupleField()
    otpOptions: list[str] = TupleField()
    otpValidFrom: datetime = TupleField()
    otpValidTo: datetime = TupleField()



import logging
from datetime import datetime

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class UserListItemTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserListItemTuple"

    #:  The unique ID of the user
    userId: str = TupleField()

    #:  The nice name of the user
    displayName: str = TupleField()

    mobile: str = TupleField()
    lastLoginDate: datetime = TupleField()
    lastLoginDeviceToken: str = TupleField()

    @property
    def userName(self) -> str:
        return self.userId

    @property
    def userTitle(self) -> str:
        return self.displayName



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserTuple(DeclarativeBase, Tuple):
    """Internal

    This table doesn't do anything

    """

    __tupleType__ = userPluginTuplePrefix + "InternalUserTuple"
    __tablename__ = "InternalUser"

    id = Column(Integer, primary_key=True, autoincrement=True)
    userName = Column(String, unique=True, nullable=False)
    userKey = Column(String, unique=True, nullable=True)
    userTitle = Column(String, unique=True, nullable=False)
    userUuid = Column(String, unique=True, nullable=False)
    # `authenticationTarget` is a custom type in postgres, see migration script
    #  43df0e05c728_added_user_import_source.py
    authenticationTarget = Column(String, nullable=False)
    # an arbitrary string of user import source
    importSource = Column(String, nullable=False)

    importHash = Column(String)

    mobile = Column(String)
    email = Column(String)

    oneTimePasscode = Column(String)
    oneTimePasscodeExpiry = Column(DateTime(timezone=True))

    groups = relationship(
        InternalGroupTuple, secondary=InternalUserGroupTuple.__table__
    )

    #: This field is ussed for the admin-app to edit the groups
    groupIds = TupleField()

    lastLoginDate = Column(DateTime(timezone=True))
    lastLoginDeviceToken = Column(String)

    __table_args__ = (Index("idx_InternalUserTable_importHash", importHash),)

    # noinspection PyMissingConstructor
    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)

    def toUserDetailTuple(self):
        return UserDetailTuple(
            userName=self.userName,
            userTitle=self.userTitle,
            userUuid=self.userUuid,
            mobile=self.mobile,
            email=self.email,
            groupNames=None,
            data=None,
            authenticationTarget=self.authenticationTarget,
            importSource=self.importSource,
        )



import logging
from datetime import datetime
from typing import List

import pytz
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.tuples.InternalGroupImportResultTuple import (
    InternalGroupImportResultTuple,
)
from peek_plugin_base.worker.CeleryApp import celeryApp
from peek_core_user.tuples.import_.ImportInternalGroupTuple import (
    ImportInternalGroupTuple,
)
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def importInternalGroups(
    self, importHash: str, groupsVortexMsg: bytes
) -> InternalGroupImportResultTuple:
    """Import Internal Groups

    :param self: A celery reference to this task
    :param importHash: An unique string of this group of items being imported.
    :param groupsVortexMsg: A vortexMsg containing the group tuples
    :returns: A list of grid keys that have been updated.
    """

    importGroups: List[ImportInternalGroupTuple] = (
        Payload().fromEncodedPayload(groupsVortexMsg).tuples
    )

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    try:
        sameCount = 0
        updates = []
        deleteIds = []
        inserts = []
        errors = []

        # This will remove duplicates
        allNames = [i.groupName for i in importGroups]

        existingGroupsByName = {
            g.groupName: g
            for g in session.query(InternalGroupTuple)
            .filter(InternalGroupTuple.userName.in_(allNames))
            .filter(InternalGroupTuple.importHash == importHash)
            .all()
        }

        for importGroup in importGroups:
            existingGroup = existingGroupsByName.pop(importGroup.grouName, None)
            if existingGroup:
                updated = False
                for fieldName in ImportInternalGroupTuple.tupleFieldNames():
                    newVal = getattr(importGroup, fieldName)
                    if getattr(existingGroup, fieldName) != newVal:
                        setattr(existingGroup, fieldName, newVal)
                        updated = True

                if updated:
                    updates.append(existingGroup)
                else:
                    sameCount += 1

            else:
                newGroup = InternalGroupTuple()

                for fieldName in ImportInternalGroupTuple.tupleFieldNames():
                    setattr(
                        newGroup, fieldName, getattr(importGroup, fieldName)
                    )

                session.add(newGroup)
                inserts.append(newGroup)

        for oldGroup in existingGroupsByName.values():
            deleteIds.append(oldGroup.id)
            session.delete(oldGroup)

        session.commit()
        logger.info(
            "Inserted %s, Updated %s, Deleted %s, Same %s, in %s",
            len(inserts),
            len(updates),
            len(deleteIds),
            sameCount,
            (datetime.now(pytz.utc) - startTime),
        )

        return InternalGroupImportResultTuple(
            addedIds=[o.id for o in inserts],
            updatedIds=[o.id for o in updates],
            deletedIds=deleteIds,
            sameCount=sameCount,
            errors=errors,
        )

    except Exception as e:
        session.rollback()
        logger.debug("Task failed, but it will retry. %s", e)
        raise self.retry(exc=e, countdown=2)

    finally:
        session.close()



import logging
from typing import List

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class InternalGroupImportResultTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "InternalGroupImportResultTuple"

    addedIds: List[int] = TupleField()
    updatedIds: List[int] = TupleField()
    deletedIds: List[int] = TupleField()
    sameCount: int = TupleField()
    errors: List[str] = TupleField()



"""removed foreign key from user logged in

Peek Plugin Database Migration Script

Revision ID: 73d65f042834
Revises: 1176bc54abf3
Create Date: 2021-04-21 15:51:50.986641

"""

# revision identifiers, used by Alembic.
revision = "73d65f042834"
down_revision = "1176bc54abf3"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_username", "UserLoggedIn")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint("uq_username", "UserLoggedIn", ["userName"])
    # ### end Alembic commands ###



"""added otp columns

Peek Plugin Database Migration Script

Revision ID: aa692f37c14d
Revises: d03db0fcb600
Create Date: 2024-02-29 15:00:43.262805

"""

# revision identifiers, used by Alembic.
revision = "aa692f37c14d"
down_revision = "2e6fbe116ff9"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        table_name="InternalUser",
        column=sa.Column("oneTimePasscode", sa.String(), nullable=True),
        schema="core_user",
    )
    op.add_column(
        table_name="InternalUser",
        column=sa.Column(
            "oneTimePasscodeExpiry", sa.DateTime(timezone=True), nullable=True
        ),
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column(
        table_name="InternalUser",
        column_name="oneTimePasscode",
        schema="core_user",
    )
    op.drop_column(
        table_name="InternalUser",
        column_name="oneTimePasscodeExpiry",
        schema="core_user",
    )
    # ### end Alembic commands ###



import logging

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_user._private.storage.Setting import (
    globalSetting,
    FIELD_SHOW_LOGIN_AS_LIST,
    FIELD_SHOW_VEHICLE_INPUT,
)
from peek_core_user._private.tuples.UserLoginUiSettingTuple import (
    UserLoginUiSettingTuple,
)
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)


class UserLoginUiSettingTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Deferred:
        tuple_ = UserLoginUiSettingTuple()

        dbSession = self._dbSessionCreator()
        try:
            tuple_.showUsersAsList = globalSetting(
                dbSession, FIELD_SHOW_LOGIN_AS_LIST
            )
            tuple_.showVehicleInput = globalSetting(
                dbSession, FIELD_SHOW_VEHICLE_INPUT
            )

        finally:
            dbSession.close()

        payloadEnvelope = Payload(
            filt=filt, tuples=[tuple_]
        ).makePayloadEnvelope()
        vortexMsg = payloadEnvelope.toVortexMsg()
        return vortexMsg



import logging
import uuid
from datetime import datetime
from typing import List

import pytz
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import subqueryload
from txcelery.defer import DeferrableTask

from peek_plugin_base.worker import CeleryDbConn
from peek_core_user._private.server.controller.PasswordUpdateController import (
    PasswordUpdateController,
)
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.tuples.InternalUserImportResultTuple import (
    InternalUserImportResultTuple,
)
from peek_plugin_base.worker.CeleryApp import celeryApp
from peek_core_user.tuples.import_.ImportInternalUserTuple import (
    ImportInternalUserTuple,
)
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


@DeferrableTask
@celeryApp.task(bind=True)
def importInternalUsers(
    self, importHash: str, usersVortexMsg: bytes
) -> InternalUserImportResultTuple:
    """Import Internal Users

    :param self: A celery reference to this task
    :param importHash: An unique string of this group of items being imported.
    :param usersVortexMsg: A vortexMsg containing the user tuples
    :returns: A list of grid keys that have been updated.
    """

    importUsers: List[ImportInternalUserTuple] = (
        Payload().fromEncodedPayload(usersVortexMsg).tuples
    )

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    try:
        same = []
        updates = []
        deleteIds = []
        inserts = []
        errors = []

        # This will remove duplicates
        allUuids = [i.userUuid for i in importUsers]

        if not allUuids:
            existingUsersByUuid = {}

        else:
            existingUsersByUuid = {
                g.userUuid: g
                for g in session.query(InternalUserTuple)
                .filter(InternalUserTuple.userUuid.in_(allUuids))
                .filter(InternalUserTuple.importHash == importHash)
                .options(subqueryload(InternalUserTuple.groups))
                .all()
            }

        groupsByName = {
            g.groupName: g for g in session.query(InternalGroupTuple).all()
        }

        for importUser in importUsers:
            try:
                existingUser = existingUsersByUuid.pop(
                    importUser.userUuid, None
                )
                if existingUser:
                    _updateUser(
                        existingUser, groupsByName, importUser, same, updates
                    )

                else:
                    _insertUser(
                        session, groupsByName, importUser, importHash, inserts
                    )

                session.commit()

            except IntegrityError as e:
                errors.append(str(e))
                session.rollback()

        for oldUser in existingUsersByUuid.values():
            deleteIds.append(oldUser.id)
            session.delete(oldUser)

        session.commit()

        logger.info(
            "Inserted %s, Updated %s, Deleted %s, Same %s, in %s",
            len(inserts),
            len(updates),
            len(deleteIds),
            len(same),
            (datetime.now(pytz.utc) - startTime),
        )

        return InternalUserImportResultTuple(
            addedIds=[o.id for o in inserts],
            updatedIds=[o.id for o in updates],
            deletedIds=deleteIds,
            sameCount=len(same),
            errors=errors,
        )

    except Exception as e:
        session.rollback()
        logger.debug("Task failed, but it will retry. %s", e)
        raise self.retry(exc=e, countdown=2)

    finally:
        session.close()


def _insertUser(session, groupsByName, importUser, importHash, inserts):
    newUser = InternalUserTuple()
    newUser.importHash = importHash

    for fieldName in ImportInternalUserTuple.tupleFieldNames():
        value = getattr(importUser, fieldName)

        if fieldName == "userName":
            # fill 'userKey' as well
            setattr(newUser, "userKey", value.lower())
        setattr(newUser, fieldName, value)

    if importUser.groupKeys is not None:
        for groupKey in importUser.groupKeys:
            newUser.groups.append(groupsByName[groupKey])

    newUser.password = PasswordUpdateController.hashPass(str(uuid.uuid4()))

    newUser.userKey = newUser.userKey.lower()
    newUser.userName = newUser.userName.lower()

    session.add(newUser)
    inserts.append(newUser)


def _updateUser(existingUser, groupsByName, importUser, same, updates):
    excludeFieldNames = ("groupKeys", "password", "userUuid")

    copyFields = filter(
        lambda f: f not in excludeFieldNames,
        ImportInternalUserTuple.tupleFieldNames(),
    )

    updated = False
    for fieldName in copyFields:
        newVal = getattr(importUser, fieldName)
        existingVal = getattr(existingUser, fieldName)
        if existingVal != newVal:
            if existingVal and not newVal:
                """Don't wipe out values if they already exist"""

            else:
                setattr(existingUser, fieldName, newVal)
                # update userKey as well when 'userName' changes
                if fieldName == "userName":
                    setattr(existingUser, "userKey", newVal.lower())
                updated = True

    # The password is an optional field
    if importUser.password is not None:
        existingUser.password = PasswordUpdateController.hashPass(
            importUser.password
        )
        updated = True

    # If there are NONE groups, then don't make any changes
    if importUser.groupKeys is not None:
        linkedGroupNames = set([g.groupName for g in existingUser.groups])
        addGroups = set(importUser.groupKeys) - linkedGroupNames
        removeGroups = linkedGroupNames - set(importUser.groupKeys)

        for addGroup in addGroups:
            existingUser.groups.append(groupsByName[addGroup])

        for removeGroup in removeGroups:
            existingUser.groups.remove(groupsByName[removeGroup])

        updated = updated or bool(addGroups) or bool(removeGroups)

    if updated:
        updates.append(existingUser)
    else:
        same.append(existingUser)



from typing import Type

from peek_plugin_base.agent.PluginAgentEntryHookABC import (
    PluginAgentEntryHookABC,
)
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)
from peek_plugin_base.server.PluginLogicEntryHookABC import (
    PluginLogicEntryHookABC,
)
from peek_plugin_base.worker.PluginWorkerEntryHookABC import (
    PluginWorkerEntryHookABC,
)

__version__ = '4.1.6'


def peekFieldEntryHook() -> Type[PluginClientEntryHookABC]:
    from peek_core_user._private.client.PluginClientEntryHook import (
        PluginClientEntryHook,
    )

    return PluginClientEntryHook


def peekOfficeEntryHook() -> Type[PluginClientEntryHookABC]:
    from peek_core_user._private.client.PluginClientEntryHook import (
        PluginClientEntryHook,
    )

    return PluginClientEntryHook


def peekAgentEntryHook() -> Type[PluginAgentEntryHookABC]:
    from ._private.agent.AgentEntryHook import AgentEntryHook

    return AgentEntryHook


def peekLogicEntryHook() -> Type[PluginLogicEntryHookABC]:
    from peek_core_user._private.server.PluginLogicEntryHook import (
        PluginLogicEntryHook,
    )

    return PluginLogicEntryHook


def peekWorkerEntryHook() -> Type[PluginWorkerEntryHookABC]:
    from ._private.worker.WorkerEntryHook import WorkerEntryHook

    return WorkerEntryHook



import logging

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.api.UserAdminAuthApi import UserAdminAuthApi
from peek_core_user._private.server.api.UserFieldHookApi import UserFieldHookApi
from peek_core_user._private.server.api.UserImportApi import UserImportApi
from peek_core_user._private.server.api.UserInfoApi import UserInfoApi
from peek_core_user._private.server.api.UserLoginApi import UserLoginApi
from peek_core_user._private.server.controller.AdminAuthController import (
    AdminAuthController,
)
from peek_core_user._private.server.controller.ImportController import (
    ImportController,
)
from peek_core_user._private.server.controller.LoginLogoutController import (
    LoginLogoutController,
)
from peek_core_user.server.UserApiABC import UserApiABC

logger = logging.getLogger(__name__)


class UserApi(UserApiABC):
    def __init__(self):
        self._hookApi = UserFieldHookApi()
        self._importApi = UserImportApi()
        self._infoApi = UserInfoApi()
        self._loginApi = UserLoginApi()
        self._adminAuthApi = UserAdminAuthApi()

    def setStartValues(
        self,
        dbSessionCreator,
        deviceApi: DeviceApiABC,
        importController: ImportController,
        loginLogoutController: LoginLogoutController,
        adminAuthController: AdminAuthController,
    ):
        self._importApi.setStartValues(importController=importController)

        self._infoApi.setStartValues(
            deviceApi=deviceApi, dbSessionCreator=dbSessionCreator
        )

        self._loginApi.setStartValues(
            loginLogoutController=loginLogoutController
        )

        self._adminAuthApi.setStartValues(
            adminAuthController=adminAuthController
        )

    def shutdown(self):
        self._loginApi.shutdown()
        self._importApi.shutdown()
        self._hookApi.shutdown()
        self._infoApi.shutdown()
        self._adminAuthApi.shutdown()

        self._loginApi = None
        self._importApi = None
        self._hookApi = None
        self._infoApi = None
        self._adminAuthApi = None

    @property
    def loginApi(self) -> UserLoginApi:
        return self._loginApi

    @property
    def importApi(self) -> UserImportApi:
        return self._importApi

    @property
    def fieldHookApi(self) -> UserFieldHookApi:
        return self._hookApi

    @property
    def infoApi(self) -> UserInfoApi:
        return self._infoApi

    @property
    def adminAuth(self) -> UserAdminAuthApi:
        return self._adminAuthApi



"""initial

Peek Plugin Database Migration Script

Revision ID: 3fec57d8da58
Revises: 
Create Date: 2017-07-13 22:44:36.374877

"""

# revision identifiers, used by Alembic.
revision = "3fec57d8da58"
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "UserLoggedIn",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("loggedInDateTime", sa.DateTime(), nullable=False),
        sa.Column("userId", sa.String(length=50), nullable=True),
        sa.Column("clientDeviceToken", sa.String(length=100), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("clientDeviceToken"),
        sa.UniqueConstraint("userId"),
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("UserLoggedIn", schema="core_user")
    # ### end Alembic commands ###



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column, ForeignKey
from sqlalchemy import Integer, String, Index
from sqlalchemy.orm import relationship

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from vortex.Tuple import Tuple, addTupleType, TupleField

logger = logging.getLogger(__name__)


class InternalUserPassword(DeclarativeBase):
    """Internal

    This table doesn't do anything

    """

    __tablename__ = "InternalUserPassword"

    id = Column(Integer, primary_key=True, autoincrement=True)

    userId = Column(
        Integer,
        ForeignKey("InternalUser.id", ondelete="CASCADE"),
        nullable=False,
    )
    user = relationship(InternalUserTuple)

    password = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_InternalUserPassword_userId", userId, unique=True),
    )



import logging

from twisted.internet.defer import Deferred, inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user._private.tuples.LoggedInUserStatusTuple import (
    LoggedInUserStatusTuple,
)

logger = logging.getLogger(__name__)


class LoggedInUserStatusTupleProvider(TuplesProviderABC):
    def __init__(self, dbSessionCreator, deviceApi: DeviceApiABC):
        self._dbSessionCreator = dbSessionCreator
        self._deviceApi = deviceApi

        assert isinstance(
            self._deviceApi, DeviceApiABC
        ), "We didn't get a DeviceApiABC"

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Deferred:
        tuples = yield self._loadLoggedInTuples()
        tuplesByDeviceToken = {t.deviceToken: t for t in tuples}

        deviceTokens = list(tuplesByDeviceToken)

        deviceDetails = yield self._deviceApi.deviceDetails(deviceTokens)

        for deviceDetail in deviceDetails:
            tuple_ = tuplesByDeviceToken[deviceDetail.deviceToken]
            tuple_.deviceLastOnline = deviceDetail.lastOnline
            tuple_.deviceType = deviceDetail.deviceType
            tuple_.deviceDescription = deviceDetail.description
            tuple_.deviceIsOnline = bool(
                deviceDetail.deviceStatus & deviceDetail.DEVICE_ONLINE
            )

        payloadEnvelope = yield Payload(
            filt=filt, tuples=tuples
        ).makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg

    @deferToThreadWrapWithLogger(logger)
    def _loadLoggedInTuples(self) -> Deferred:
        dbSession = self._dbSessionCreator()
        try:
            tuples = [
                LoggedInUserStatusTuple(
                    userName=u.userName,
                    userTitle=u.userTitle,
                    loginDate=u.loggedInDateTime,
                    vehicle=u.vehicle,
                    deviceToken=u.deviceToken,
                )
                for u in dbSession.query(
                    UserLoggedIn.userName,
                    UserLoggedIn.loggedInDateTime,
                    UserLoggedIn.deviceToken,
                    UserLoggedIn.vehicle,
                    InternalUserTuple.userTitle,
                ).join(
                    InternalUserTuple,
                    UserLoggedIn.userName == InternalUserTuple.userName,
                )
            ]

            return tuples

        finally:
            dbSession.close()



"""Convert userKey and userName to all lowercase for InternalUser

Peek Plugin Database Migration Script

Revision ID: d03db0fcb600
Revises: eef485ef2e0e
Create Date: 2023-02-22 11:49:39.624413

"""

# revision identifiers, used by Alembic.
revision = "d03db0fcb600"
down_revision = "eef485ef2e0e"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    op.execute(
        """DELETE FROM core_user."InternalUser"
        WHERE id in (
            SELECT id
            FROM core_user."InternalUser"
            WHERE lower("userKey") in (
                SELECT lower("userKey")
                FROM core_user."InternalUser"
                GROUP BY lower("userKey")
                HAVING count(lower("userKey")) > 1
            )
        )
        """
    )

    op.execute(
        """DELETE FROM core_user."InternalUser"
        WHERE id in (
            SELECT id
            FROM core_user."InternalUser"
            WHERE lower("userName") in (
                SELECT lower("userName")
                FROM core_user."InternalUser"
                GROUP BY lower("userName")
                HAVING count(lower("userName")) > 1
            )
        )
        """
    )

    op.execute(
        'UPDATE core_user."InternalUser" SET "userKey" = LOWER("userKey")'
    )
    op.execute(
        'UPDATE core_user."InternalUser" SET "userName" = LOWER("userName")'
    )


def downgrade():
    pass



import logging

from twisted.internet.defer import Deferred

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.server.controller.LoginLogoutController import (
    LoginLogoutController,
)
from peek_core_user.server.UserLoginApiABC import UserLoginApiABC
from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLoginAuthMethodAction import (
    UserLoginAuthMethodAction,
)
from peek_core_user.tuples.login.UserLoginOtpAction import UserLoginOtpAction

logger = logging.getLogger(__name__)


class UserLoginApi(UserLoginApiABC):
    #: A reference to the core device plugins API
    _deviceApi: DeviceApiABC

    def __init__(self):
        self._loginLogoutController = None

    def setStartValues(self, loginLogoutController: LoginLogoutController):
        self._loginLogoutController = loginLogoutController

    def shutdown(self):
        self._loginLogoutController = None

    def logout(self, logoutTuple: UserLogoutAction) -> Deferred:
        return self._loginLogoutController.logout(logoutTuple)

    def login(self, loginTuple: UserLoginAction) -> Deferred:
        return self._loginLogoutController.login(loginTuple)

    def queryUserLoginAuthMethod(
        self, userLoginAuthMethod: UserLoginAuthMethodAction
    ) -> Deferred:
        return self._loginLogoutController.queryUserLoginAuthMethod(
            userLoginAuthMethod
        )

    def loginRequestOtp(self, userLoginOtpAction: UserLoginOtpAction):
        return self._loginLogoutController.loginRequestOtp(userLoginOtpAction)



"""Added agentUri column to LdapSetting

Peek Plugin Database Migration Script

Revision ID: eef485ef2e0e
Revises: d79da9333e2a
Create Date: 2022-08-24 09:10:13.594230

"""

# revision identifiers, used by Alembic.
revision = "eef485ef2e0e"
down_revision = "d79da9333e2a"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "LdapSetting",
        sa.Column(
            "agentHost",
            sa.String(),
            nullable=True,
        ),
        schema="core_user",
    )
    op.execute(
        """
        UPDATE "core_user"."LdapSetting" SET "agentHost" = NULL;
        """
    )


def downgrade():
    op.drop_column("LdapSetting", "agentHost", schema="core_user")



import logging

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class LdapLoggedInUserTuple(Tuple):
    """Ldap Logged In User Tuple

    Aggregates the result of a successfully logged in user from LDAP

    """

    __tupleType__ = userPluginTuplePrefix + "LdapLoggedInUserTuple"

    username: str = TupleField()
    userTitle: str = TupleField()
    userUuid: str = TupleField()
    email: str = TupleField()
    ldapName: str = TupleField()
    objectSid: str = TupleField()
    ldapDomain: str = TupleField()



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class InternalGroupTuple(DeclarativeBase, Tuple):
    """Group Table

    This table contains the user plugin groups, for the internal directory.

    """

    __tupleType__ = userPluginTuplePrefix + "InternalGroupTuple"
    __tablename__ = "InternalGroup"

    id = Column(Integer, primary_key=True, autoincrement=True)
    groupName = Column(String, unique=True, nullable=False)
    groupTitle = Column(String, unique=True, nullable=False)

    importHash = Column(String)

    __table_args__ = (Index("idx_InternalGroupTable_importHash", importHash),)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



import logging

from twisted.internet.defer import inlineCallbacks, Deferred

from peek_core_user._private.server.api.UserApi import UserApi
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class GroupDetailTupleProvider(TuplesProviderABC):
    def __init__(self, ourApi: UserApi):
        self._ourApi = ourApi

        from peek_core_user.server.UserApiABC import UserApiABC

        assert isinstance(
            self._ourApi, UserApiABC
        ), "We didn't get a UserApiABC"

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Deferred:
        tuples = yield self._ourApi.infoApi.groups()

        payloadEnvelope = yield Payload(
            filt=filt, tuples=tuples
        ).makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg
